// Copyright (c) 2016, XMOS Ltd, All rights reserved
#include <xs1.h>
#include "i2c.h"
#include "usb.h"
#include "hid_mouse_extended.h"

// Accelerometer code taken from AN00181

// FXOS8700EQ register address defines
#define FXOS8700EQ_I2C_ADDR 0x1E
#define FXOS8700EQ_XYZ_DATA_CFG_REG 0x0E
#define FXOS8700EQ_CTRL_REG_1 0x2A
#define FXOS8700EQ_DR_STATUS 0x0
#define FXOS8700EQ_OUT_X_MSB 0x1
#define FXOS8700EQ_OUT_X_LSB 0x2
#define FXOS8700EQ_OUT_Y_MSB 0x3
#define FXOS8700EQ_OUT_Y_LSB 0x4
#define FXOS8700EQ_OUT_Z_MSB 0x5
#define FXOS8700EQ_OUT_Z_LSB 0x6

int read_acceleration(client interface i2c_master_if i2c, int reg) {
    i2c_regop_res_t result;
    int accel_val = 0;
    unsigned char data = 0;

    // Read MSB data
    data = i2c.read_reg(FXOS8700EQ_I2C_ADDR, reg, result);
    if (result != I2C_REGOP_SUCCESS) {
      return 0;
    }

    accel_val = data << 2;

    // Read LSB data
    data = i2c.read_reg(FXOS8700EQ_I2C_ADDR, reg+1, result);
    if (result != I2C_REGOP_SUCCESS) {
      return 0;
    }

    accel_val |= (data >> 6);

    if (accel_val & 0x200) {
      accel_val -= 1023;
    }

    return accel_val;
}

void get_accelerometer_data(client interface i2c_master_if i2c, int &x, int &y) {
   i2c_regop_res_t result;
   char status_data = 0;
   // Wait for valid accelerometer data
   do {
      status_data = i2c.read_reg(FXOS8700EQ_I2C_ADDR, FXOS8700EQ_DR_STATUS, result);
   } while (!status_data & 0x08);

   // Read x and y axis values
   x = read_acceleration(i2c, FXOS8700EQ_OUT_X_MSB);
   y = read_acceleration(i2c, FXOS8700EQ_OUT_Y_MSB);
}

void init_accelerometer(client interface i2c_master_if i2c) {
    i2c_regop_res_t result;

    // Configure FXOS8700EQ
    result = i2c.write_reg(FXOS8700EQ_I2C_ADDR, FXOS8700EQ_XYZ_DATA_CFG_REG, 0x01);
    if (result != I2C_REGOP_SUCCESS) {
      return;
    }

    // Enable FXOS8700EQ
    result = i2c.write_reg(FXOS8700EQ_I2C_ADDR, FXOS8700EQ_CTRL_REG_1, 0x01);
    if (result != I2C_REGOP_SUCCESS) {
      return;
    }
}

// Button access
// PORT_4E connected to the 2 Buttons
on tile[0]: port p_button = XS1_PORT_4E;
#define DEBOUNCE_TIME   (XS1_TIMER_HZ/50)
#define BUTTON_PRESSED  0x00

/* Function to get button state (0 or 1)*/
int get_button_state(int button_id)
{
    int button_val;
    p_button :> button_val;
    button_val = (button_val >> button_id) & (0x01);
    return button_val;
}

/* Checks if a button is pressed */
int is_button_pressed(int button_id)
{
    if(get_button_state(button_id) == BUTTON_PRESSED) {
        /* Wait for debounce and check again */
        delay_ticks(DEBOUNCE_TIME);
        if(get_button_state(button_id) == BUTTON_PRESSED) {
            return 1; /* Yes button is pressed */
        }
    }
    /* No button press */
    return 0;
}

// Mouse input interface, reads from accelerometer and buttons
void mouse_input(server interface mouse_input_if mi_if, client interface i2c_master_if i2c) {
    init_accelerometer(i2c);
    while (1) {
        select {
            case mi_if.get_input_data() -> struct mouse_input_data data:
            get_accelerometer_data(i2c, data.x, data.y);
            data.button_left = is_button_pressed(1);
            data.button_right = is_button_pressed(0);
            break;
        }
    }
}

// HID report buffer used by endpoint 0
extern unsigned char g_reportBuffer[4];

/*
 * This function responds to the HID requests
 */
void hid_mouse_extended(chanend chan_ep_hid, client interface mouse_input_if mi_if)
{
    XUD_ep ep_hid = XUD_InitEp(chan_ep_hid, XUD_EPTYPE_INT);

    while (1)
    {
      /* Unsafe region so we can use shared memory. */
      unsafe {
        char * unsafe p_reportBuffer = g_reportBuffer;
        int x, y;
        struct mouse_input_data mi_data;

        // Initialise the HID
        p_reportBuffer[0] = 0;  // Buttons
        p_reportBuffer[1] = 0;  // X axis
        p_reportBuffer[2] = 0;  // Y axis

        // Read from mouse input interface
        mi_data = mi_if.get_input_data();

        // Set up HID x axis value report
        if (mi_data.x > 10 || mi_data.x < -10) {
          p_reportBuffer[1] = mi_data.x >> 2;
        }

        // Set up HID y axis value report
        if (mi_data.y > 10 || mi_data.y < -10) {
          p_reportBuffer[2] = (mi_data.y * -1) >> 2;
        }

        // Set up HID button value report
        p_reportBuffer[0] |= mi_data.button_left;
        p_reportBuffer[0] |= mi_data.button_right << 1;

        /* Send the buffer off to the host.  Note this will return when complete */
        XUD_SetBuffer(ep_hid, (char *) p_reportBuffer, 4);
      }
    }
}


